import bgheatmaps as bgh


values = {
    "facial motor nucleus": 3.5,
    "anterior trigeminal motor region": 0.5,
    "locus coeruleus": -2,
    "dorsal thalamus": -1,
    "ventral habenula": -4,
    "caudal hypothalamus": 2.5,
    "pituitary": 4,
    "posterior tuberculum": -3,
    "preoptic region": 3,
    "olfactory bulb": 5,
    "medial tegmentum": -2,
    "tectal neuropil": -4,
    "oculomotor nucleus": 1,
    "tegmentum": -0.5,
    "inferior olive": -0.75,
    "inferior medulla oblongata": 1.5,
    "cerebellum": 0.5,
}

f = bgh.heatmap(
    values,
    position=None,
    orientation="sagittal",  # 'frontal' or 'sagittal', or 'horizontal' or a tuple (x,y,z)
    thickness=1000,
    atlas_name="mpin_zfish_1um",
    format="2D",
    title="zebra fish heatmap",
).show(xlabel="AP (μm)", ylabel="DV (μm)")



import pandas as pd

from bg_atlasapi.bg_atlas import BrainGlobeAtlas
from brainrender._io import load_mesh_from_file

import bgheatmaps as bgh

"""
    This exmaple shows how to go from cell counts per region (e.g. as outputted by
    cellfinder) to a plot showing the density (count/volume) of cells in each brain region
"""

# get the number of cells for each region
data = pd.read_hdf("examples/cell_counts_example.h5")
cell_counts = data.groupby("region").count()

# get regions two levels up the hierarchy
atlas = BrainGlobeAtlas("allen_mouse_25um")
parent_regions = [
    atlas.get_structure_ancestors(r)[-2] for r in cell_counts.index
]

cell_counts["parent_region"] = parent_regions
cell_counts = cell_counts.groupby("parent_region").sum()


# get regions' volume
volumes = []
for region in cell_counts.index:
    obj_file = str(atlas.meshfile_from_structure(region))
    mesh = load_mesh_from_file(obj_file)
    volumes.append(mesh.volume())

# get the density (num cells / volume)
cell_counts["volume"] = volumes
cell_counts["density"] = cell_counts["x"] / cell_counts["volume"]
cell_counts = cell_counts.loc[cell_counts.density > 5 * 1e-9]

print(cell_counts)


f = bgh.heatmap(
    cell_counts.density.to_dict(),
    position=(8000, 5000, 5000,),
    orientation="frontal",  # or 'sagittal', or 'horizontal' or a tuple (x,y,z)
    title="cell density",
    format="2D",
    cmap="Reds",
).show()



import bgheatmaps as bgh


values = {
    "2Ssp": 3.5,
    "3Sp": 0.5,
    "IB": -2,
    "5Sp": -1,
    "7Sp": -4,
    "LDCom": 2.5,
    "10Sp": 4,
    "D": -3,
    "ICl": -4,
    "6Sp": 4,
    "gr": 8,
    "vf": -1,
    "rs": -4,
    "LSp": 7,
    "dcs": -5,
}


f = bgh.heatmap(
    values,
    position=1000,
    orientation="sagittal",  # 'frontal' or 'sagittal', or 'horizontal' or a tuple (x,y,z)
    thickness=1000,
    atlas_name="allen_cord_20um",
    format="2D",
).show()



import bgheatmaps as bgh

"""
    This example shows how to use the `plan` functionality to visualize the 
    position of the planes used to 'slice' the brain for heatmap visualization.
"""

regions = [
    "TH",
    "RSP",
    "AI",
    "SS",
    "MO",
    "PVZ",
    "LZ",
    "VIS",
    "AUD",
    "RHP",
    "STR",
    "CB",
    "FRP",
    "HIP",
    "PA",
]


planner = bgh.plan(
    regions,
    position=(8000, 5000, 5000,),
    orientation="frontal",  # or 'sagittal', or 'horizontal' or a tuple (x,y,z)
    thickness=2000,  # thickness of the slices used for rendering (in microns)
    arrow_scale=750,
)
planner.show()



from typing import Union
import numpy as np
from vedo import Arrow, Plane, Sphere
from rich.table import Table
from rich.panel import Panel
from rich import print

from myterial import pink_dark, blue_dark, amber_lighter, grey, amber, orange

from bgheatmaps.heatmaps import heatmap

from brainrender import settings

settings.BACKGROUND_COLOR = amber_lighter
settings.ROOT_COLOR = grey


def print_plane(name: str, plane: Plane, color: str):
    """
        Prints nicely formatted information about a plane
    """

    def fmt_array(x: np.ndarray) -> str:
        return str(tuple([round(v, 2) for v in x]))

    # create a table to display the vertices posittion
    vert_tb = Table(box=None)
    vert_tb.add_column(style=f"{amber}", justify="right")
    vert_tb.add_column(style="white")

    for i in range(4):
        vert_tb.add_row(f"({i})", fmt_array(plane.mesh.points()[i]))

    tb = Table(box=None)
    tb.add_column(style=f"bold {orange}", justify="right")
    tb.add_column(style="white")

    tb.add_row("center point: ", fmt_array(plane.mesh.center))
    tb.add_row("norm: ", str(tuple(plane.mesh.normal)))
    tb.add_row("Vertices: ", vert_tb)

    print(
        Panel.fit(
            tb, title=f"[white bold]{name}", style=color, title_align="left"
        )
    )


class plan(heatmap):
    def __init__(
        self,
        regions: Union[dict, list],
        position: Union[list, tuple, np.ndarray],
        orientation: Union[str, tuple] = "frontal",
        thickness: float = 10,
        arrow_scale: float = 10,
        **kwargs,
    ):
        self.arrow_scale = arrow_scale

        if isinstance(regions, list):
            regions = {r: 1 for r in regions}

        super().__init__(
            regions,
            position=position,
            orientation=orientation,
            thickness=thickness,
            format="3D",
            **kwargs,
        )

        # print planes information
        print_plane("Plane 0", self.slicer.plane0, blue_dark)
        print_plane("Plane 1", self.slicer.plane1, pink_dark)

    def show(self):
        """
            Renders the hetamap visualization as a 3D scene in brainrender.
        """
        self.scene.root._mesh.alpha(0.3)

        # show sliced brain regions
        self.slicer.show_plane_intersection(
            self.scene, self.regions_meshes, self.scene.root
        )

        # add slicing planes and their norms
        for plane, color, alpha in zip(
            (self.slicer.plane0, self.slicer.plane1),
            (blue_dark, pink_dark),
            (0.8, 0.3),
        ):
            plane.alpha(alpha).color(color)

            self.scene.add(plane, transform=False)
            self.scene.add(
                Arrow(
                    plane.center,
                    np.array(plane.center)
                    + np.array(plane.mesh.normal) * self.arrow_scale,
                    c=color,
                ),
                transform=False,
            )

            self.scene.add(
                Sphere(plane.center, r=plane.width / 125, c="k"),
                transform=False,
            )

        self.scene.render(interactive=self.interactive, zoom=self.zoom)
        return self.scene


if __name__ == "__main__":
    regions = dict(  # scalar values for each region
        TH=1,
        RSP=0.2,
        AI=0.4,
        SS=-3,
        MO=2.6,
        PVZ=-4,
        LZ=-3,
        VIS=2,
        AUD=0.3,
        RHP=-0.2,
        STR=0.5,
        CB=0.5,
        FRP=-1.7,
        HIP=3,
        PA=-4,
    )

    plan(
        regions,
        position=(
            8000,
            5000,
            5000,
        ),  # position of the center point of the plane
        orientation="frontal",  # or 'sagittal', or 'horizontal' or a tuple (x,y,z)
        thickness=2000,  # thickness of the slices used for rendering (in microns)
        arrow_scale=750,
    ).show()



from typing import List, Dict, Union, Optional
from vedo import Plane
import numpy as np

from brainrender.actor import Actor
from brainrender.scene import Scene


def get_ax_idx(orientation: str) -> int:
    """
        Given a named orientation get the idx
        of the axis orthogonal to the plane,
    """
    if orientation == "frontal":
        return 0
    elif orientation == "sagittal":
        return 2
    elif orientation == "horizontal":
        return 1
    else:
        raise ValueError(f'Orientation "{orientation}" not recognized')


class Slicer:
    def __init__(
        self,
        position: Optional[Union[list, tuple, np.ndarray, float]],
        orientation: Union[str, tuple],
        thickness: float,
        root: Actor,
    ):
        """
            Computes the position of two planes given a point (position) and an orientation (named orientation or 
            3D vector) + thickness (spacing between the two planes)
        """
        if position is None:
            position = root.centerOfMass()

        if isinstance(position, (float, int)):
            if isinstance(orientation, str):
                pval = position
                position = root.centerOfMass()
                position[get_ax_idx(orientation)] = pval
            else:
                raise ValueError(
                    "When a single float value is passed for position, the orientation should be one of the named orientations values"
                )

        position = np.array(position)
        position[2] = -position[2]

        if isinstance(orientation, str):
            axidx = get_ax_idx(orientation)

            # get the two points the plances are cenered at
            shift = np.zeros(3)
            shift[axidx] -= thickness
            p1 = position - shift

            # get the two planes
            norm0, norm1 = np.zeros(3), np.zeros(3)
            norm0[axidx] = 1
            norm1[axidx] = -1
        else:
            orientation = np.array(orientation)

            p1 = position + orientation * thickness  # type: ignore

            norm0 = orientation  # type: ignore
            norm1 = -orientation  # type: ignore

        # get the length of the largest dimension of the atlas
        bounds = root.bounds()
        length = max(
            bounds[1] - bounds[0],
            bounds[3] - bounds[2],
            bounds[5] - bounds[4],
        )
        length += length / 3

        self.plane0 = Actor(
            Plane(pos=position, normal=norm0, sx=length, sy=length),
            name=f"Plane at {position} norm: {norm0}",
            br_class="plane",
        )
        self.plane0.width = length

        self.plane1 = Actor(
            Plane(pos=p1, normal=norm1, sx=length, sy=length),
            name=f"Plane at {p1} norm: {norm1}",
            br_class="plane",
        )
        self.plane1.width = length

    def get_structures_slice_coords(self, regions: List[Actor], root: Actor):
        """
        It computes the intersection between the first slice plane and all 
        user given brain regions,
        returning the coordinates of each region as a set of XY (i.e. in the plane's
        coordinates system) coordinates
        """
        regions = regions + [root]

        pts = self.plane0.points() - self.plane0.points()[0]
        v = pts[1] / np.linalg.norm(pts[1])
        w = pts[2] / np.linalg.norm(pts[2])

        M = np.vstack([v, w]).T  # 3 x 2

        projected: Dict[str, np.ndarray] = {}
        for n, actor in enumerate(regions):
            # get region/plane intersection
            intersection = self.plane0.intersectWith(
                actor._mesh.triangulate()
            )  # points: (N x 3)

            if not intersection.points().shape[0]:
                continue  # no intersection

            pieces = intersection.splitByConnectivity()
            for piece_n, piece in enumerate(pieces):

                # sort coordinates
                points = piece.join(reset=True).points()

                projected[actor.name + f"_segment_{piece_n}"] = points @ M

        # get output coordinates
        coordinates: Dict[str, List[np.ndarray]] = dict()

        for region in [r.name for r in regions]:
            coordinates[region] = [
                v for k, v in projected.items() if region in k
            ]

        return projected, coordinates

    def show_plane_intersection(
        self, scene: Scene, regions: List[Actor], root: Actor
    ):
        """
            Slices regions' meshjes with plane0 and adds the resulting intersection
            to the brainrender scene.
        """
        for region in regions + [root]:
            intersection = self.plane0.intersectWith(region._mesh)

            if len(intersection.points()):
                scene.add(intersection, transform=False)

            if region.name != "root":
                scene.remove(region)

    def slice_scene(self, scene: Scene, regions: List[Actor]):
        """
            Slices the meshes in a 3D brainrender scene using the gien planes
        """
        # slice the scene
        for n, plane in enumerate((self.plane0, self.plane1)):
            scene.slice(plane, actors=regions, close_actors=True)

        scene.slice(self.plane0, actors=scene.root, close_actors=False)


def get_structures_slice_coords(
    regions: List[str],
    position: Union[list, tuple, np.ndarray],
    orientation: Union[str, tuple] = "frontal",
    atlas_name: Optional[str] = None,
) -> Dict[str, List[np.ndarray]]:
    """
        Given a list of regions name and a set of plane parameters, it returns 
        the coordinates of the plane/regions' intersection in the plane's coordinates
    """

    scene = Scene(atlas_name=atlas_name)
    if len(regions) == 1:
        regions_actors = [scene.add_brain_region(*regions)]
    else:
        regions_actors = scene.add_brain_region(*regions)

    slicer = Slicer(position, orientation, 100, scene.root)

    structures_coords = slicer.get_structures_slice_coords(
        regions_actors, scene.root
    )[1]
    scene.close()
    return structures_coords



from vedo.colors import colorMap as map_color
from typing import Optional, Union, Dict, Tuple
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib as mpl

from myterial import grey_darker

from brainrender import Scene
from brainrender import settings, cameras
from brainrender.atlas import Atlas

from bgheatmaps.slicer import Slicer

# Set settings for heatmap visualization
settings.SHOW_AXES = False
settings.SHADER_STYLE = "cartoon"
settings.ROOT_ALPHA = 0.3
settings.ROOT_COLOR = grey_darker


def check_values(values: dict, atlas: Atlas) -> Tuple[float, float]:
    """
    Checks that the passed heatmap values meet two criteria:
        - keys should be acronyms of brainregions
        - values should be numbers
    """
    for k, v in values.items():
        if not isinstance(v, (float, int)):
            raise ValueError(
                f'Heatmap values should be floats, not: {type(v)} for entry "{k}"'
            )

        if k not in atlas.lookup_df.acronym.values:
            raise ValueError(f'Region name "{k}" not recognized')

    vmax, vmin = max(values.values()), min(values.values())
    return vmax, vmin


class heatmap:
    def __init__(
        self,
        values: Dict[str, float],
        position: Union[list, tuple, np.ndarray],
        orientation: Union[str, tuple] = "frontal",
        title: Optional[str] = None,
        cmap: str = "Reds",
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        format: str = "3D",  # 3D -> brainrender, 2D -> matplotlib
        # brainrender, 3D HM specific
        thickness: float = 10,
        interactive: bool = True,
        zoom: Optional[float] = None,
        atlas_name: Optional[str] = None,
        **kwargs,
    ):
        # store arguments
        self.values = values
        self.format = format
        self.orientation = orientation
        self.interactive = interactive
        self.zoom = zoom
        self.title = title
        self.cmap = cmap

        # create a scene
        self.scene = Scene(
            atlas_name=atlas_name,
            title=title,
            title_color=grey_darker,
            **kwargs,
        )

        # prep colors range
        self.prepare_colors(values, cmap, vmin, vmax)

        # add regions to the brainrender scene
        self.scene.add_brain_region(*self.values.keys())

        self.regions_meshes = [
            r
            for r in self.scene.get_actors(br_class="brain region")
            if r.name != "root"
        ]

        # prepare slicer object
        self.slicer = Slicer(position, orientation, thickness, self.scene.root)

    def prepare_colors(
        self,
        values: dict,
        cmap: str,
        vmin: Optional[float],
        vmax: Optional[float],
    ):
        # get brain regions colors
        _vmax, _vmin = check_values(values, self.scene.atlas)
        if _vmax == _vmin:
            _vmin = _vmax * 0.5

        vmin = vmin or _vmin
        vmax = vmax or _vmax
        self.vmin, self.vmax = vmin, vmax

        self.colors = {
            r: list(map_color(v, name=cmap, vmin=vmin, vmax=vmax))
            for r, v in values.items()
        }
        self.colors["root"] = grey_darker

    def show(self, **kwargs) -> Union[Scene, plt.Figure]:
        """
            Creates a 2D plot or 3D rendering of the heatmap
        """
        if self.format == "3D":
            self.slicer.slice_scene(self.scene, self.regions_meshes)
            view = self.render(**kwargs)
        else:
            view = self.plot(**kwargs)
        return view

    def render(self, **kwargs) -> Scene:
        """
            Renders the hetamap visualization as a 3D scene in brainrender.
        """

        # set brain regions colors
        for region, color in self.colors.items():
            if region == "root":
                continue

            self.scene.get_actors(br_class="brain region", name=region)[
                0
            ].color(color)

        # set camera position and render
        if isinstance(self.orientation, str):
            if self.orientation == "sagittal":
                camera = cameras.sagittal_camera2
            elif self.orientation == "horizontal":
                camera = "top"
            else:
                camera = self.orientation
        else:
            self.orientation = np.array(self.orientation)
            com = self.slicer.plane0.centerOfMass()
            camera = {
                "pos": com - self.orientation * 2 * np.linalg.norm(com),
                "viewup": (0, -1, 0),
                "clippingRange": (19531, 40903),
            }

        self.scene.render(
            camera=camera, interactive=self.interactive, zoom=self.zoom
        )
        return self.scene

    def plot(
        self,
        show_legend: bool = False,
        xlabel: str = "μm",
        ylabel: str = "μm",
        **kwargs,
    ) -> plt.Figure:
        """
            Plots the heatmap in 2D using matplotlib
        """
        self.scene.close()
        projected, _ = self.slicer.get_structures_slice_coords(
            self.regions_meshes, self.scene.root
        )

        f, ax = plt.subplots(figsize=(9, 9))
        for r, coords in projected.items():
            name, segment = r.split("_segment_")
            ax.fill(
                coords[:, 0],
                coords[:, 1],
                color=self.colors[name],
                label=name if segment == "0" and name != "root" else None,
                lw=1,
                ec="k",
                zorder=-1 if name == "root" else None,
                alpha=0.3 if name == "root" else None,
            )

        # make colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        # cmap = mpl.cm.cool
        norm = mpl.colors.Normalize(vmin=self.vmin, vmax=self.vmax)
        f.colorbar(mpl.cm.ScalarMappable(norm=norm, cmap=self.cmap), cax=cax)

        # style axes
        ax.invert_yaxis()
        ax.axis("equal")
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.set(title=self.title)

        if isinstance(self.orientation, str) or np.sum(self.orientation) == 1:
            # orthogonal projection
            ax.set(xlabel=xlabel, ylabel=ylabel)

        if show_legend:
            ax.legend()
        plt.show()

        return f


if __name__ == "__main__":
    values = dict(  # scalar values for each region
        TH=1,
        RSP=0.2,
        AI=0.4,
        SS=-3,
        MO=2.6,
        PVZ=-4,
        LZ=-3,
        VIS=2,
        AUD=0.3,
        RHP=-0.2,
        STR=0.5,
        CB=0.5,
        FRP=-1.7,
        HIP=3,
        PA=-4,
    )

    heatmap(
        values,
        position=(8000, 5000, 5000,),
        orientation=(
            1,
            1,
            1,
        ),  # or 'sagittal', or 'horizontal' or a tuple (x,y,z)
        thickness=250,  # thickness of the slices used for rendering (in microns)
        title="frontal",
        vmin=-5,
        vmax=3,
        format="3D",
    ).show()



import bgheatmaps as bgh

"""
    This example shows how to get the coordinates 
    of selected brain regions in the slicing plane
"""

regions = [
    "TH",
    "RSP",
    "AI",
    "SS",
    "MO",
    "PVZ",
    "LZ",
    "VIS",
    "AUD",
    "RHP",
    "STR",
    "CB",
    "FRP",
    "HIP",
    "PA",
]


coordinates = bgh.get_structures_slice_coords(
    regions,
    position=(8000, 5000, 5000,),
    orientation="frontal",  # 'frontal' or 'sagittal', or 'horizontal' or a tuple (x,y,z)
)

print(coordinates)



from setuptools import setup, find_namespace_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = ["brainrender", "matplotlib", "numpy", "myterial", "rich"]

setup(
    name="bgheatmap",
    version="0.0.1",
    description="Brain regions heatmaps in brainrender",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires=">=3.6",
    packages=find_namespace_packages(exclude=("control")),
    include_package_data=True,
    url="",
    author="Federico Claudi",
    zip_safe=False,
    entry_points={"console_scripts": []},
)



import bgheatmaps as bgh

"""
    This example shows how to use visualize a heatmap in 2D
"""

values = dict(  # scalar values for each region
    TH=1,
    RSP=0.2,
    AI=0.4,
    SS=-3,
    MO=2.6,
    PVZ=-4,
    LZ=-3,
    VIS=2,
    AUD=0.3,
    RHP=-0.2,
    STR=0.5,
    CB=0.5,
    FRP=-1.7,
    HIP=3,
    PA=-4,
)


f = bgh.heatmap(
    values,
    position=5000,  # when using a named orientation you can pass a single value!
    orientation="frontal",  # 'frontal' or 'sagittal', or 'horizontal' or a tuple (x,y,z)
    title="horizontal view",
    vmin=-5,
    vmax=3,
    format="2D",
).show()



import bgheatmaps as bgh

from matplotlib import pyplot as plt
import matplotlib.patheffects as path_effects

from bg_atlasapi import BrainGlobeAtlas


"""
    This example shows how to use bgheatmaps to get the contours of 'sliced'
    brain regions to overlay them over histological data.
"""

atlas_name = "mpin_zfish_1um"  # name of the atlas

dors_vent_slice_pos = 130  # position of the slice in um

# Cut coordinates from meshes:
coords_dict = bgh.get_structures_slice_coords(
    ["tectum", "midbrain"],
    orientation="horizontal",
    atlas_name="mpin_zfish_1um",
    position=(0, dors_vent_slice_pos, 0),
)
_ = coords_dict.pop("root")  # we won't plot the root mask

# Get atlas slice:
atlas = BrainGlobeAtlas(atlas_name)
ref_slice_idx = int(dors_vent_slice_pos / atlas.resolution[0])
confocal_slice = atlas.reference[:, ref_slice_idx, :]

# plot:
f, ax = plt.subplots(figsize=(6, 9))
plt.imshow(confocal_slice, cmap="gray_r")
for struct_name, contours in coords_dict.items():
    for cont in contours:
        plt.fill(cont[:, 0], -cont[:, 1], lw=1, fc="none", ec="k")

        # write outlined text to label each brain region
        txt = plt.text(
            cont[:, 0].mean(),
            -cont[:, 1].mean(),
            struct_name[:4] + ".",
            ha="center",
            va="center",
            fontsize=12,
        )
        txt.set_path_effects(
            [path_effects.withStroke(linewidth=6, foreground="white",)]
        )


ax.set(xlabel="Left - Right pos (µm)", ylabel="Post. - Ant. pos (µm)")
plt.show()
